@@ -1,7 +1,7 @@
 /*
  * threads.c	request threading support
  *
- * Version:	$Id: threads.c,v 1.140 2007/11/23 13:46:51 aland Exp $
+ * Version:	$Id: d4de4ff436e4b485aeed76fc3c76a5daf966a143 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id: threads.c,v 1.140 2007/11/23 13:46:51 aland Exp $")
+RCSID("$Id: d4de4ff436e4b485aeed76fc3c76a5daf966a143 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/rad_assert.h>
@@ -79,7 +79,7 @@ RCSID("$Id: threads.c,v 1.140 2007/11/23 13:46:51 aland Exp $")
  *  the current thread.
  *
  *  pthread_id     pthread id
- *  thread_num     server thread number, 1...number of threads
+ *  thread_id     server thread number, 1...number of threads
  *  semaphore     used to block the thread until a request comes in
  *  status        is the thread running or exited?
  *  request_count the number of requests that this thread has handled
@@ -89,7 +89,7 @@ typedef struct THREAD_HANDLE {
 	struct THREAD_HANDLE *prev;
 	struct THREAD_HANDLE *next;
 	pthread_t            pthread_id;
-	int                  thread_num;
+	int                  thread_id;
 	int                  status;
 	unsigned int         request_count;
 	time_t               timestamp;
@@ -120,9 +120,10 @@ typedef struct THREAD_POOL {
 	THREAD_HANDLE *head;
 	THREAD_HANDLE *tail;
 
-	int total_threads;
 	int active_threads;	/* protected by queue_mutex */
-	int max_thread_num;
+	int exited_threads;
+	int total_threads;
+	int max_thread_id;
 	int start_threads;
 	int max_threads;
 	int min_spare_threads;
@@ -131,10 +132,13 @@ typedef struct THREAD_POOL {
 	unsigned long request_count;
 	time_t time_last_spawned;
 	int cleanup_delay;
+	int stop_flag;
 	int spawn_flag;
 
+#ifdef WNOHANG
 	pthread_mutex_t	wait_mutex;
 	fr_hash_table_t *waiters;
+#endif
 
 	/*
 	 *	All threads wait on this semaphore, for requests
@@ -149,13 +153,13 @@ typedef struct THREAD_POOL {
 
 	int		max_queue_size;
 	int		num_queued;
-	int		can_read_detail;
 	fr_fifo_t	*fifo[NUM_FIFOS];
 } THREAD_POOL;
 
 static THREAD_POOL thread_pool;
 static int pool_initialized = FALSE;
 static time_t last_cleaned = 0;
+static time_t almost_now = 0;
 
 static void thread_pool_manage(time_t now);
 
@@ -235,7 +239,7 @@ static int setup_ssl_mutexes(void)
 }
 #endif
 
-
+#ifdef WNOHANG
 /*
  *	We don't want to catch SIGCHLD for a host of reasons.
  *
@@ -262,12 +266,13 @@ static void reap_children(void)
 	pthread_mutex_lock(&thread_pool.wait_mutex);
 
 	do {
+	retry:
 		pid = waitpid(0, &status, WNOHANG);
 		if (pid <= 0) break;
 
 		mytf.pid = pid;
 		tf = fr_hash_table_finddata(thread_pool.waiters, &mytf);
-		if (!tf) continue;
+		if (!tf) goto retry;
 
 		tf->status = status;
 		tf->exited = 1;
@@ -275,6 +280,9 @@ static void reap_children(void)
 
 	pthread_mutex_unlock(&thread_pool.wait_mutex);
 }
+#else
+#define reap_children()
+#endif /* WNOHANG */
 
 /*
  *	Add a request to the list of waiting requests.
@@ -286,20 +294,46 @@ static int request_enqueue(REQUEST *request, RAD_REQUEST_FUNP fun)
 {
 	request_queue_t *entry;
 
+	/*
+	 *	If we haven't checked the number of child threads
+	 *	in a while, OR if the thread pool appears to be full,
+	 *	go manage it.
+	 */
+	if ((last_cleaned < request->timestamp) ||
+	    (thread_pool.active_threads == thread_pool.total_threads) ||
+	    (thread_pool.exited_threads > 0)) {
+		thread_pool_manage(request->timestamp);
+	}
+
 	pthread_mutex_lock(&thread_pool.queue_mutex);
 
 	thread_pool.request_count++;
 
 	if (thread_pool.num_queued >= thread_pool.max_queue_size) {
+		int complain = FALSE;
+		time_t now;
+		static time_t last_complained = 0;
+
+		now = time(NULL);
+		if (last_complained != now) {
+			last_complained = now;
+			complain = TRUE;
+		}
+		    
 		pthread_mutex_unlock(&thread_pool.queue_mutex);
 
 		/*
 		 *	Mark the request as done.
 		 */
-		radlog(L_ERR, "!!! ERROR !!! The server is blocked: discarding new request %d", request->number);
+		if (complain) {
+			radlog(L_ERR, "Something is blocking the server.  There are %d packets in the queue, waiting to be processed.  Ignoring the new request.", thread_pool.max_queue_size);
+		}
 		request->child_state = REQUEST_DONE;
 		return 0;
 	}
+	request->child_state = REQUEST_QUEUED;
+	request->component = "<core>";
+	request->module = "<queue>";
 
 	entry = rad_malloc(sizeof(*entry));
 	entry->request = request;
@@ -316,15 +350,6 @@ static int request_enqueue(REQUEST *request, RAD_REQUEST_FUNP fun)
 		return 0;
 	}
 
-	/*
-	 *	We've added an entry that didn't come from the detail
-	 *	file.  Note that the child thread should signal the
-	 *	main worker thread again when the queue becomes empty.
-	 */
-	if (request->listener->type != RAD_LISTEN_DETAIL) {
-		thread_pool.can_read_detail = FALSE;
-	}
-
 	thread_pool.num_queued++;
 
 	pthread_mutex_unlock(&thread_pool.queue_mutex);
@@ -348,6 +373,7 @@ static int request_enqueue(REQUEST *request, RAD_REQUEST_FUNP fun)
  */
 static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 {
+	int blocked;
 	RAD_LISTEN_TYPE i, start;
 	request_queue_t *entry;
 
@@ -375,6 +401,9 @@ static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 		entry = fr_fifo_pop(thread_pool.fifo[i]);
 		rad_assert(entry != NULL);
 		entry->request->child_state = REQUEST_DONE;
+		thread_pool.num_queued--;
+		free(entry);
+		entry = NULL;
 	}
 
 	start = 0;
@@ -402,11 +431,15 @@ static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 	*request = entry->request;
 	*fun = entry->fun;
 	free(entry);
+	entry = NULL;
 
 	rad_assert(*request != NULL);
 	rad_assert((*request)->magic == REQUEST_MAGIC);
 	rad_assert(*fun != NULL);
 
+	(*request)->component = "<core>";
+	(*request)->module = "<thread>";
+
 	/*
 	 *	If the request has sat in the queue for too long,
 	 *	kill it.
@@ -416,10 +449,29 @@ static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 	 *	have acknowledged it as "done".
 	 */
 	if ((*request)->master_state == REQUEST_STOP_PROCESSING) {
+		(*request)->module = "<done>";
 		(*request)->child_state = REQUEST_DONE;
 		goto retry;
 	}
 
+	/*
+	 *	Produce messages for people who have 10 million rows
+	 *	in a database, without indexes.
+	 */
+	rad_assert(almost_now != 0);
+	blocked = almost_now - (*request)->timestamp;
+	if (blocked < 5) {
+		blocked = 0;
+	} else {
+		static time_t last_complained = 0;
+		
+		if (last_complained != almost_now) {
+			last_complained = almost_now;
+		} else {
+			blocked = 0;
+		}
+	}
+
 	/*
 	 *	The thread is currently processing a request.
 	 */
@@ -427,10 +479,14 @@ static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 
 	pthread_mutex_unlock(&thread_pool.queue_mutex);
 
+	if (blocked) {
+		radlog(L_ERR, "Request %u has been waiting in the processing queue for %d seconds.  Check that all databases are running properly!",
+		       (*request)->number, blocked);
+	}
+
 	return 1;
 }
 
-
 /*
  *	The main thread handler for requests.
  *
@@ -445,13 +501,11 @@ static void *request_handler_thread(void *arg)
 	 *	Loop forever, until told to exit.
 	 */
 	do {
-		int can_read_detail;
-
 		/*
 		 *	Wait to be signalled.
 		 */
 		DEBUG2("Thread %d waiting to be assigned a request",
-		       self->thread_num);
+		       self->thread_id);
 	re_wait:
 		if (sem_wait(&thread_pool.semaphore) != 0) {
 			/*
@@ -460,15 +514,28 @@ static void *request_handler_thread(void *arg)
 			 *	text.
 			 */
 			if (errno == EINTR) {
-				DEBUG2("Re-wait %d", self->thread_num);
+				DEBUG2("Re-wait %d", self->thread_id);
 				goto re_wait;
 			}
 			radlog(L_ERR, "Thread %d failed waiting for semaphore: %s: Exiting\n",
-			       self->thread_num, strerror(errno));
+			       self->thread_id, strerror(errno));
 			break;
 		}
 
-		DEBUG2("Thread %d got semaphore", self->thread_num);
+		DEBUG2("Thread %d got semaphore", self->thread_id);
+
+#ifdef HAVE_OPENSSL_ERR_H
+ 		/*
+		 *	Clear the error queue for the current thread.
+		 */
+		ERR_clear_error ();
+#endif
+
+		/*
+		 *	The server is exiting.  Don't dequeue any
+		 *	requests.
+		 */
+		if (thread_pool.stop_flag) break;
 
 		/*
 		 *	Try to grab a request from the queue.
@@ -478,14 +545,16 @@ static void *request_handler_thread(void *arg)
 		 */
 		if (!request_dequeue(&self->request, &fun)) continue;
 
-		self->request->child_pid = self->pthread_id;
+		self->request->thread_id = self->thread_id;
 		self->request_count++;
 
 		DEBUG2("Thread %d handling request %d, (%d handled so far)",
-		       self->thread_num, self->request->number,
+		       self->thread_id, self->request->number,
 		       self->request_count);
 
+		self->request->module = "";
 		radius_handle_request(self->request, fun);
+		self->request = NULL;
 
 		/*
 		 *	Update the active threads.
@@ -493,32 +562,21 @@ static void *request_handler_thread(void *arg)
 		pthread_mutex_lock(&thread_pool.queue_mutex);
 		rad_assert(thread_pool.active_threads > 0);
 		thread_pool.active_threads--;
-
-		/*
-		 *	If we're not currently allowed to read the
-		 *	detail file, AND there are no requests queued,
-		 *	THEN signal the main worker thread that
-		 *	there's at least one waiting thread (us) who
-		 *	can accept a packet from the detail file.
-		 */
-		can_read_detail = FALSE;
-		if (!thread_pool.can_read_detail &&
-		    (thread_pool.num_queued == 0)) {
-			can_read_detail = TRUE;
-		}
-
 		pthread_mutex_unlock(&thread_pool.queue_mutex);
 
 		/*
-		 *	Do this out of the lock to be nice to everyone.
+		 *	If the thread has handled too many requests, then make it
+		 *	exit.
 		 */
-		if (can_read_detail) {
-			radius_signal_self(RADIUS_SIGNAL_SELF_DETAIL);
+		if ((thread_pool.max_requests_per_thread > 0) &&
+		    (self->request_count >= thread_pool.max_requests_per_thread)) {
+			DEBUG2("Thread %d handled too many requests",
+			       self->thread_id);
+			break;
 		}
-
 	} while (self->status != THREAD_CANCELLED);
 
-	DEBUG2("Thread %d exiting...", self->thread_num);
+	DEBUG2("Thread %d exiting...", self->thread_id);
 
 #ifdef HAVE_OPENSSL_ERR_H
 	/*
@@ -529,6 +587,10 @@ static void *request_handler_thread(void *arg)
 	ERR_remove_state(0);
 #endif
 
+	pthread_mutex_lock(&thread_pool.queue_mutex);
+	thread_pool.exited_threads++;
+	pthread_mutex_unlock(&thread_pool.queue_mutex);
+
 	/*
 	 *  Do this as the LAST thing before exiting.
 	 */
@@ -552,7 +614,7 @@ static void delete_thread(THREAD_HANDLE *handle)
 
 	rad_assert(handle->request == NULL);
 
-	DEBUG2("Deleting thread %d", handle->thread_num);
+	DEBUG2("Deleting thread %d", handle->thread_id);
 
 	prev = handle->prev;
 	next = handle->next;
@@ -593,7 +655,6 @@ static THREAD_HANDLE *spawn_thread(time_t now)
 {
 	int rcode;
 	THREAD_HANDLE *handle;
-	pthread_attr_t attr;
 
 	/*
 	 *	Ensure that we don't spawn too many threads.
@@ -610,44 +671,32 @@ static THREAD_HANDLE *spawn_thread(time_t now)
 	memset(handle, 0, sizeof(THREAD_HANDLE));
 	handle->prev = NULL;
 	handle->next = NULL;
-	handle->pthread_id = NO_SUCH_CHILD_PID;
-	handle->thread_num = thread_pool.max_thread_num++;
+	handle->thread_id = thread_pool.max_thread_id++;
 	handle->request_count = 0;
 	handle->status = THREAD_RUNNING;
 	handle->timestamp = time(NULL);
 
 	/*
-	 *	Initialize the thread's attributes to detached.
-	 *
-	 *	We could call pthread_detach() later, but if the thread
-	 *	exits between the create & detach calls, it will need to
-	 *	be joined, which will never happen.
-	 */
-	pthread_attr_init(&attr);
-	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
-
-	/*
-	 *	Create the thread detached, so that it cleans up it's
-	 *	own memory when it exits.
+	 *	Create the thread joinable, so that it can be cleaned up
+	 *	using pthread_join().
 	 *
 	 *	Note that the function returns non-zero on error, NOT
 	 *	-1.  The return code is the error, and errno isn't set.
 	 */
-	rcode = pthread_create(&handle->pthread_id, &attr,
+	rcode = pthread_create(&handle->pthread_id, 0,
 			request_handler_thread, handle);
 	if (rcode != 0) {
 		radlog(L_ERR, "Thread create failed: %s",
 		       strerror(rcode));
 		return NULL;
 	}
-	pthread_attr_destroy(&attr);
 
 	/*
 	 *	One more thread to go into the list.
 	 */
 	thread_pool.total_threads++;
 	DEBUG2("Thread spawned new child %d. Total threads in pool: %d",
-			handle->thread_num, thread_pool.total_threads);
+			handle->thread_id, thread_pool.total_threads);
 
 	/*
 	 *	Add the thread handle to the tail of the thread pool list.
@@ -689,6 +738,7 @@ int total_active_threads(void)
 }
 
 
+#ifdef WNOHANG
 static uint32_t pid_hash(const void *data)
 {
 	const thread_fork_t *tf = data;
@@ -703,6 +753,7 @@ static int pid_cmp(const void *one, const void *two)
 
 	return (a->pid - b->pid);
 }
+#endif
 
 /*
  *	Allocate the thread pool, and seed it with an initial number
@@ -710,7 +761,7 @@ static int pid_cmp(const void *one, const void *two)
  *
  *	FIXME: What to do on a SIGHUP???
  */
-int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
+int thread_pool_init(CONF_SECTION *cs, int *spawn_flag)
 {
 	int		i, rcode;
 	CONF_SECTION	*pool_cf;
@@ -718,51 +769,50 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
 
 	now = time(NULL);
 
-	/*
-	 *	We're not spawning new threads, don't do
-	 *	anything.
-	 */
-	if (!spawn_flag) return 0;
-
-	/*
-	 *	After a SIGHUP, we don't over-write the previous values.
-	 */
-	if (!pool_initialized) {
-		/*
-		 *	Initialize the thread pool to some reasonable values.
-		 */
-		memset(&thread_pool, 0, sizeof(THREAD_POOL));
-		thread_pool.head = NULL;
-		thread_pool.tail = NULL;
-		thread_pool.total_threads = 0;
-		thread_pool.max_thread_num = 1;
-		thread_pool.cleanup_delay = 5;
-		thread_pool.spawn_flag = spawn_flag;
-
-		if ((pthread_mutex_init(&thread_pool.wait_mutex,NULL) != 0)) {
-			radlog(L_ERR, "FATAL: Failed to initialize wait mutex: %s",
-			       strerror(errno));
-			return -1;
-		}
-
-		/*
-		 *	Create the hash table of child PID's
-		 */
-		thread_pool.waiters = fr_hash_table_create(pid_hash,
-							     pid_cmp,
-							     free);
-		if (!thread_pool.waiters) {
-			radlog(L_ERR, "FATAL: Failed to set up wait hash");
-			return -1;
-		}
-	}
+	rad_assert(spawn_flag != NULL);
+	rad_assert(*spawn_flag == TRUE);
+	rad_assert(pool_initialized == FALSE); /* not called on HUP */
 
 	pool_cf = cf_subsection_find_next(cs, NULL, "thread");
-	if (!pool_cf) {
-		radlog(L_ERR, "FATAL: Attempting to start in multi-threaded mode with no thread configuration in radiusd.conf");
+	if (!pool_cf) *spawn_flag = FALSE;
+
+	/*
+	 *	Initialize the thread pool to some reasonable values.
+	 */
+	memset(&thread_pool, 0, sizeof(THREAD_POOL));
+	thread_pool.head = NULL;
+	thread_pool.tail = NULL;
+	thread_pool.total_threads = 0;
+	thread_pool.max_thread_id = 1;
+	thread_pool.cleanup_delay = 5;
+	thread_pool.stop_flag = 0;
+	thread_pool.spawn_flag = *spawn_flag;
+	
+	/*
+	 *	Don't bother initializing the mutexes or
+	 *	creating the hash tables.  They won't be used.
+	 */
+	if (!*spawn_flag) return 0;
+	
+#ifdef WNOHANG
+	if ((pthread_mutex_init(&thread_pool.wait_mutex,NULL) != 0)) {
+		radlog(L_ERR, "FATAL: Failed to initialize wait mutex: %s",
+		       strerror(errno));
 		return -1;
 	}
 
+	/*
+	 *	Create the hash table of child PID's
+	 */
+	thread_pool.waiters = fr_hash_table_create(pid_hash,
+						   pid_cmp,
+						   free);
+	if (!thread_pool.waiters) {
+		radlog(L_ERR, "FATAL: Failed to set up wait hash");
+		return -1;
+	}
+#endif
+
 	if (cf_section_parse(pool_cf, NULL, thread_config) < 0) {
 		return -1;
 	}
@@ -776,6 +826,10 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
 		thread_pool.max_spare_threads = 1;
 	if (thread_pool.max_spare_threads < thread_pool.min_spare_threads)
 		thread_pool.max_spare_threads = thread_pool.min_spare_threads;
+	if ((thread_pool.max_queue_size < 2) || (thread_pool.max_queue_size > 1048576)) {
+		radlog(L_ERR, "FATAL: max_queue_size value must be in range 2-1048576");
+		return -1;
+	}
 
 	/*
 	 *	The pool has already been initialized.  Don't spawn
@@ -807,7 +861,7 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
 	 *	Allocate multiple fifos.
 	 */
 	for (i = 0; i < RAD_LISTEN_MAX; i++) {
-		thread_pool.fifo[i] = fr_fifo_create(65536, NULL);
+		thread_pool.fifo[i] = fr_fifo_create(thread_pool.max_queue_size, NULL);
 		if (!thread_pool.fifo[i]) {
 			radlog(L_ERR, "FATAL: Failed to set up request fifo");
 			return -1;
@@ -843,6 +897,40 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
 }
 
 
+/*
+ *	Stop all threads in the pool.
+ */
+void thread_pool_stop(void)
+{
+	int i;
+	int total_threads;
+	THREAD_HANDLE *handle;
+	THREAD_HANDLE *next;
+
+	/*
+	 *	Set pool stop flag.
+	 */
+	thread_pool.stop_flag = 1;
+
+	/*
+	 *	Wakeup all threads to make them see stop flag.
+	 */
+	total_threads = thread_pool.total_threads;
+	for (i = 0; i != total_threads; i++) {
+		sem_post(&thread_pool.semaphore);
+	}
+
+	/*
+	 *	Join and free all threads.
+	 */
+	for (handle = thread_pool.head; handle; handle = next) {
+		next = handle->next;
+		pthread_join(handle->pthread_id, NULL);
+		delete_thread(handle);
+	}
+}
+
+
 /*
  *	Assign a new request to a free thread.
  *
@@ -851,7 +939,7 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
  */
 int thread_pool_addrequest(REQUEST *request, RAD_REQUEST_FUNP fun)
 {
-	time_t now = request->timestamp;
+	almost_now = request->timestamp;
 
 	/*
 	 *	We've been told not to spawn threads, so don't.
@@ -859,12 +947,14 @@ int thread_pool_addrequest(REQUEST *request, RAD_REQUEST_FUNP fun)
 	if (!thread_pool.spawn_flag) {
 		radius_handle_request(request, fun);
 
+#ifdef WNOHANG
 		/*
 		 *	Requests that care about child process exit
 		 *	codes have already either called
 		 *	rad_waitpid(), or they've given up.
 		 */
 		wait(NULL);
+#endif
 		return 1;
 	}
 
@@ -878,9 +968,9 @@ int thread_pool_addrequest(REQUEST *request, RAD_REQUEST_FUNP fun)
 	 *	in a while, OR if the thread pool appears to be full,
 	 *	go manage it.
 	 */
-	if ((last_cleaned < now) ||
+	if ((last_cleaned < almost_now) ||
 	    (thread_pool.active_threads == thread_pool.total_threads)) {
-		thread_pool_manage(now);
+		thread_pool_manage(almost_now);
 	}
 
 	return 1;
@@ -899,6 +989,25 @@ static void thread_pool_manage(time_t now)
 	THREAD_HANDLE *handle, *next;
 	int active_threads;
 
+	/*
+	 *	Loop over the thread pool, deleting exited threads.
+	 */
+	for (handle = thread_pool.head; handle; handle = next) {
+		next = handle->next;
+
+		/*
+		 *	Maybe we've asked the thread to exit, and it
+		 *	has agreed.
+		 */
+		if (handle->status == THREAD_EXITED) {
+			pthread_join(handle->pthread_id, NULL);
+			delete_thread(handle);
+			pthread_mutex_lock(&thread_pool.queue_mutex);
+			thread_pool.exited_threads--;
+			pthread_mutex_unlock(&thread_pool.queue_mutex);
+		}
+	}
+
 	/*
 	 *	We don't need a mutex lock here, as we're reading
 	 *	active_threads, and not modifying it.  We want a close
@@ -923,9 +1032,14 @@ static void thread_pool_manage(time_t now)
 	/*
 	 *	If there are too few spare threads.  Go create some more.
 	 */
-	if (spare < thread_pool.min_spare_threads) {
+	if ((thread_pool.total_threads < thread_pool.max_threads) &&
+	    (spare < thread_pool.min_spare_threads)) {
 		total = thread_pool.min_spare_threads - spare;
 
+		if ((total + thread_pool.total_threads) > thread_pool.max_threads) {
+			total = thread_pool.max_threads - thread_pool.total_threads;
+		}
+
 		DEBUG2("Threads: Spawning %d spares", total);
 
 		/*
@@ -950,21 +1064,6 @@ static void thread_pool_manage(time_t now)
 	}
 	last_cleaned = now;
 
-	/*
-	 *	Loop over the thread pool, deleting exited threads.
-	 */
-	for (handle = thread_pool.head; handle; handle = next) {
-		next = handle->next;
-
-		/*
-		 *	Maybe we've asked the thread to exit, and it
-		 *	has agreed.
-		 */
-		if (handle->status == THREAD_EXITED) {
-			delete_thread(handle);
-		}
-	}
-
 	/*
 	 *	Only delete the spare threads if sufficient time has
 	 *	passed since we last created one.  This helps to minimize
@@ -1016,27 +1115,6 @@ static void thread_pool_manage(time_t now)
 		}
 	}
 
-	/*
-	 *	If the thread has handled too many requests, then make it
-	 *	exit.
-	 */
-	if (thread_pool.max_requests_per_thread > 0) {
-		for (handle = thread_pool.head; handle; handle = next) {
-			next = handle->next;
-
-			/*
-			 *	Not handling a request, but otherwise
-			 *	live, we can kill it.
-			 */
-			if ((handle->request == NULL) &&
-			    (handle->status == THREAD_RUNNING) &&
-			    (handle->request_count > thread_pool.max_requests_per_thread)) {
-				handle->status = THREAD_CANCELLED;
-				sem_post(&thread_pool.semaphore);
-			}
-		}
-	}
-
 	/*
 	 *	Otherwise everything's kosher.  There are not too few,
 	 *	or too many spare threads.  Exit happily.
@@ -1045,6 +1123,7 @@ static void thread_pool_manage(time_t now)
 }
 
 
+#ifdef WNOHANG
 /*
  *	Thread wrapper for fork().
  */
@@ -1063,6 +1142,7 @@ pid_t rad_fork(void)
 	/*
 	 *	Fork & save the PID for later reaping.
 	 */
+	pthread_mutex_lock(&thread_pool.wait_mutex);
 	child_pid = fork();
 	if (child_pid > 0) {
 		int rcode;
@@ -1073,15 +1153,20 @@ pid_t rad_fork(void)
 
 		tf->pid = child_pid;
 
-		pthread_mutex_lock(&thread_pool.wait_mutex);
 		rcode = fr_hash_table_insert(thread_pool.waiters, tf);
-		pthread_mutex_unlock(&thread_pool.wait_mutex);
 
 		if (!rcode) {
 			radlog(L_ERR, "Failed to store PID, creating what will be a zombie process %d",
 			       (int) child_pid);
+			free(tf);
 		}
 	}
+	/*
+	 *	Do not unlock in child process
+	 */
+	if(child_pid != 0 ) {
+		pthread_mutex_unlock(&thread_pool.wait_mutex);
+	}
 
 	/*
 	 *	Return whatever we were told.
@@ -1133,6 +1218,11 @@ pid_t rad_waitpid(pid_t pid, int *status)
 
 	return 0;
 }
+#else
+/*
+ *	No rad_fork or rad_waitpid
+ */
+#endif
 
 void thread_pool_lock(void)
 {
@@ -1143,10 +1233,19 @@ void thread_pool_unlock(void)
 {
 	pthread_mutex_unlock(&thread_pool.queue_mutex);
 }
-#else
-int thread_pool_addrequest(REQUEST *request, RAD_REQUEST_FUNP fun)
+
+void thread_pool_queue_stats(int *array)
 {
-	radius_handle_request(request, fun);
-	return 1;
+	int i;
+
+	if (pool_initialized) {
+		for (i = 0; i < RAD_LISTEN_MAX; i++) {
+			array[i] = fr_fifo_num_elements(thread_pool.fifo[i]);
+		}
+	} else {
+		for (i = 0; i < RAD_LISTEN_MAX; i++) {
+			array[i] = 0;
+		}
+	}
 }
 #endif /* HAVE_PTHREAD_H */
