@@ -1,7 +1,7 @@
 /*
  * rlm_pap.c
  *
- * Version:  $Id: 8ef2152df423e227697a76700c95a838775e6fe5 $
+ * Version:  $Id: 6ac2f9ac22660c9dd6b123e0d80a7e003f12c7fb $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id: 8ef2152df423e227697a76700c95a838775e6fe5 $")
+RCSID("$Id: 6ac2f9ac22660c9dd6b123e0d80a7e003f12c7fb $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -247,15 +247,32 @@ static int base64_decode (const char *src, uint8_t *dst)
 static void normify(REQUEST *request, VALUE_PAIR *vp, size_t min_length)
 {
 	size_t decoded;
-	uint8_t buffer[64];
+	uint8_t buffer[256];
+	char raw[sizeof(vp->vp_strvalue) + 1];
+	char *value;
 
 	if (min_length >= sizeof(buffer)) return; /* paranoia */
+	/*
+	 *	fr_hex2bin and base64_decode don't deal well with non
+	 *	\0 terminated buffers.
+	 */
+	if (vp->type == PW_TYPE_OCTETS) {
+		if (vp->length > sizeof(raw)) return;
+
+		memcpy(raw, vp->vp_octets, vp->length);
+		raw[vp->length] = '\0';
+		value = raw;
+	} else if (vp->type == PW_TYPE_STRING) {
+		value = vp->vp_strvalue;
+	} else {
+		return;
+	}
 
 	/*
 	 *	Hex encoding.
 	 */
 	if (vp->length >= (2 * min_length)) {
-		decoded = fr_hex2bin(vp->vp_strvalue, buffer, vp->length >> 1);
+		decoded = fr_hex2bin(value, buffer, sizeof(buffer));
 		if (decoded == (vp->length >> 1)) {
 			RDEBUG2("Normalizing %s from hex encoding", vp->name);
 			memcpy(vp->vp_octets, buffer, decoded);
@@ -269,7 +286,7 @@ static void normify(REQUEST *request, VALUE_PAIR *vp, size_t min_length)
 	 *	and we want to avoid division...
 	 */
 	if (((vp->length * 3) >= ((min_length * 4))) &&
-	    ((decoded = base64_decode(vp->vp_strvalue, buffer)) > 0) &&
+	    ((decoded = base64_decode(value, buffer)) > 0) &&
 	    (decoded >= min_length)) {
 		RDEBUG2("Normalizing %s from base64 encoding", vp->name);
 		memcpy(vp->vp_octets, buffer, decoded);
@@ -371,7 +388,7 @@ static int pap_authorize(void *instance, REQUEST *request)
 			new_vp = radius_paircreate(request,
 						   &request->config_items,
 						   attr, PW_TYPE_STRING);
-			
+
 			/*
 			 *	The data after the '}' may be binary,
 			 *	so we copy it via memcpy.
