@@ -1,7 +1,7 @@
  /*
  * rlm_perl.c
  *
- * Version:    $Id$
+ * Version:    $Id: 9fe9d3d2ff60eb4c2a4cb0d18f3b919fd68f3ea0 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: 9fe9d3d2ff60eb4c2a4cb0d18f3b919fd68f3ea0 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -69,9 +69,16 @@ typedef struct perl_inst {
 	char	*func_pre_proxy;
 	char	*func_post_proxy;
 	char	*func_post_auth;
+#ifdef WITH_COA
+	char	*func_recv_coa;
+	char	*func_send_coa;
+#endif
 	char	*xlat_name;
 	char	*perl_flags;
 	PerlInterpreter *perl;
+	pthread_key_t	*thread_key;
+
+	pthread_mutex_t clone_mutex;
 } PERL_INST;
 /*
  *	A mapping of configuration file names to internal variables.
@@ -105,6 +112,12 @@ static const CONF_PARSER module_config[] = {
 	  offsetof(PERL_INST,func_post_proxy), NULL, "post_proxy"},
 	{ "func_post_auth", PW_TYPE_STRING_PTR,
 	  offsetof(PERL_INST,func_post_auth), NULL, "post_auth"},
+#ifdef WITH_COA
+	{ "func_recv_coa", PW_TYPE_STRING_PTR,
+	  offsetof(PERL_INST,func_recv_coa), NULL, "recv_coa"},
+	{ "func_send_coa", PW_TYPE_STRING_PTR,
+	  offsetof(PERL_INST,func_send_coa), NULL, "send_coa"},
+#endif
 	{ "perl_flags", PW_TYPE_STRING_PTR,
 	  offsetof(PERL_INST,perl_flags), NULL, NULL},
 	{ "func_start_accounting", PW_TYPE_STRING_PTR,
@@ -138,12 +151,7 @@ static void **rlm_perl_get_handles(pTHX)
 	AV *modules = get_av(dl_modules, FALSE);
 	void **handles;
 
-	if (!librefs) {
-		radlog(L_ERR,
-		   "Could not get @%s for unloading.\n",
-		   dl_librefs);
-		return NULL;
-	}
+	if (!librefs) return NULL;
 
 	if (!(AvFILL(librefs) >= 0)) {
 		return NULL;
@@ -194,7 +202,6 @@ static void rlm_perl_close_handles(void **handles)
 
 static void rlm_perl_destruct(PerlInterpreter *perl)
 {
-	char **orig_environ = NULL;
 	dTHXa(perl);
 
 	PERL_SET_CONTEXT(perl);
@@ -216,10 +223,6 @@ static void rlm_perl_destruct(PerlInterpreter *perl)
 
 	perl_destruct(perl);
 	perl_free(perl);
-
-	if (orig_environ) {
-		environ = orig_environ;
-	}
 }
 
 static void rlm_destroy_perl(PerlInterpreter *perl)
@@ -230,29 +233,24 @@ static void rlm_destroy_perl(PerlInterpreter *perl)
 	PERL_SET_CONTEXT(perl);
 
 	handles = rlm_perl_get_handles(aTHX);
+	if (handles) rlm_perl_close_handles(handles);
 	rlm_perl_destruct(perl);
-	rlm_perl_close_handles(handles);
 }
 
-static pthread_key_t  rlm_perl_key;
-static pthread_once_t rlm_perl_once = PTHREAD_ONCE_INIT;
-
 /* Create Key */
-static void rlm_perl_make_key(void)
+static void rlm_perl_make_key(pthread_key_t *key)
 {
-	pthread_key_create(&rlm_perl_key, rlm_destroy_perl);
+	pthread_key_create(key, rlm_destroy_perl);
 }
 
-static PerlInterpreter *rlm_perl_clone(PerlInterpreter *perl)
+static PerlInterpreter *rlm_perl_clone(PerlInterpreter *perl, pthread_key_t *key)
 {
 	PerlInterpreter *interp;
 	UV clone_flags = 0;
 
 	PERL_SET_CONTEXT(perl);
 
-	pthread_once(&rlm_perl_once, rlm_perl_make_key);
-
-	interp = pthread_getspecific(rlm_perl_key);
+	interp = pthread_getspecific(*key);
 	if (interp) return interp;
 
 	interp = perl_clone(perl, clone_flags);
@@ -268,22 +266,12 @@ static PerlInterpreter *rlm_perl_clone(PerlInterpreter *perl)
 	PERL_SET_CONTEXT(aTHX);
     	rlm_perl_clear_handles(aTHX);
 
-	pthread_setspecific(rlm_perl_key, interp);
-
-	fprintf(stderr, "GOT CLONE %d %p\n", pthread_self(), interp);
+	pthread_setspecific(*key, interp);
 
 	return interp;
 }
 #endif
 
-static void xs_init(pTHX)
-{
-	char *file = __FILE__;
-
-	/* DynaLoader is a special case */
-	newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
-
-}
 /*
  *
  * This is wrapper for radlog
@@ -312,6 +300,16 @@ static XS(XS_radiusd_radlog)
        XSRETURN_NO;
 }
 
+static void xs_init(pTHX)
+{
+	char *file = __FILE__;
+
+	/* DynaLoader is a special case */
+	newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
+
+	newXS("radiusd::radlog",XS_radiusd_radlog, "rlm_perl");
+}
+
 /*
  * The xlat function
  */
@@ -320,8 +318,8 @@ static size_t perl_xlat(void *instance, REQUEST *request, char *fmt, char *out,
 {
 
 	PERL_INST	*inst= (PERL_INST *) instance;
-	PerlInterpreter *perl;
-	char		params[1024], *ptr, *tmp;
+	char		params[1024], *tmp;
+	char const	*p, *q;
 	int		count;
 	size_t		ret = 0;
 	STRLEN		n_a;
@@ -334,51 +332,57 @@ static size_t perl_xlat(void *instance, REQUEST *request, char *fmt, char *out,
 		return 0;
 	}
 
-#ifndef WITH_ITHREADS
-	perl = inst->perl;
+#ifdef USE_ITHREADS
+	pthread_mutex_lock(&inst->clone_mutex);
+
+	PerlInterpreter *interp;
+
+	interp = rlm_perl_clone(inst->perl,inst->thread_key);
+	{
+	  dTHXa(interp);
+	  PERL_SET_CONTEXT(interp);
+	}
+
+	pthread_mutex_unlock(&inst->clone_mutex);
 #else
-	perl = rlm_perl_clone(inst->perl);
-	{
-	  dTHXa(perl);
-	}
+	PERL_SET_CONTEXT(inst->perl);
 #endif
-	PERL_SET_CONTEXT(perl);
+
 	{
-	dSP;
-	ENTER;SAVETMPS;
+		dSP;
+		ENTER;SAVETMPS;
 
-	ptr = strtok(params, " ");
+		PUSHMARK(SP);
 
-	PUSHMARK(SP);
+		p = fmt;
+		while ((q = strchr(p, ' '))) {
+			XPUSHs(sv_2mortal(newSVpv(p, p - q)));
 
-	while (ptr != NULL) {
-		XPUSHs(sv_2mortal(newSVpv(ptr,0)));
-		ptr = strtok(NULL, " ");
-	}
+			p = q + 1;
+		}
 
-	PUTBACK;
+		PUTBACK;
 
-	count = call_pv(inst->func_xlat, G_SCALAR | G_EVAL);
+		count = call_pv(inst->func_xlat, G_SCALAR | G_EVAL);
 
-	SPAGAIN;
-	if (SvTRUE(ERRSV)) {
-		radlog(L_ERR, "rlm_perl: perl_xlat exit %s\n",
-		       SvPV(ERRSV,n_a));
-		POPs ;
-	} else if (count > 0) {
-		tmp = POPp;
-		strlcpy(out, tmp, freespace);
-		ret = strlen(out);
+		SPAGAIN;
+		if (SvTRUE(ERRSV)) {
+			RDEBUG("ERROR: Exit %s", SvPV(ERRSV,n_a));
+			(void)POPs;
+		} else if (count > 0) {
+			tmp = POPp;
+			strlcpy(out, tmp, freespace);
+			ret = strlen(out);
 
-		radlog(L_DBG,"rlm_perl: Len is %d , out is %s freespace is %d",
-		       ret, out,freespace);
-	}
+			RDEBUG("Len is %zu , out is %s freespace is %zu", ret, out, freespace);
+		}
 
-	PUTBACK ;
-	FREETMPS ;
-	LEAVE ;
+		PUTBACK ;
+		FREETMPS ;
+		LEAVE ;
 
 	}
+
 	return ret;
 }
 /*
@@ -407,10 +411,13 @@ static int perl_instantiate(CONF_SECTION *conf, void **instance)
 	HV		*rad_request_proxy_reply_hv;
 	AV		*end_AV;
 
-	char *embed[4];
+	char **embed;
+	char **envp = NULL;
 	const char *xlat_name;
 	int exitstatus = 0, argc=0;
 
+	embed = rad_malloc(4 * sizeof(char *));
+	memset(embed, 0, 4 *sizeof(char *));
 	/*
 	 *	Set up a storage area for instance data
 	 */
@@ -422,11 +429,23 @@ static int perl_instantiate(CONF_SECTION *conf, void **instance)
 	 *	fail.
 	 */
 	if (cf_section_parse(conf, inst, module_config) < 0) {
+		free(embed);
 		free(inst);
 		return -1;
 	}
 
+	/*
+	 *	Create pthread key. This key will be stored in instance
+	 */
 
+#ifdef USE_ITHREADS
+	pthread_mutex_init(&inst->clone_mutex, NULL);
+
+	inst->thread_key = rad_malloc(sizeof(*inst->thread_key));
+	memset(inst->thread_key,0,sizeof(*inst->thread_key));
+
+	rlm_perl_make_key(inst->thread_key);
+#endif
 	embed[0] = NULL;
 	if (inst->perl_flags) {
 		embed[1] = inst->perl_flags;
@@ -439,9 +458,12 @@ static int perl_instantiate(CONF_SECTION *conf, void **instance)
 		argc = 3;
 	}
 
+	PERL_SYS_INIT3(&argc, &embed, &envp);
 #ifdef USE_ITHREADS
 	if ((inst->perl = perl_alloc()) == NULL) {
 		radlog(L_DBG, "rlm_perl: No memory for allocating new perl !");
+		free(embed);
+		free(inst);
 		return (-1);
 	}
 
@@ -455,6 +477,8 @@ static int perl_instantiate(CONF_SECTION *conf, void **instance)
 #else
 	if ((inst->perl = perl_alloc()) == NULL) {
 		radlog(L_ERR, "rlm_perl: No memory for allocating new perl !");
+		free(embed);
+		free(inst);
 		return -1;
 	}
 
@@ -474,13 +498,13 @@ static int perl_instantiate(CONF_SECTION *conf, void **instance)
 		exitstatus = perl_run(inst->perl);
 	} else {
 		radlog(L_ERR,"rlm_perl: perl_parse failed: %s not found or has syntax errors. \n", inst->module);
+		free(embed);
+		free(inst);
 		return (-1);
 	}
 
 	PL_endav = end_AV;
 
-        newXS("radiusd::radlog",XS_radiusd_radlog, "rlm_perl.c");
-
 	rad_reply_hv = newHV();
 	rad_check_hv = newHV();
 	rad_config_hv = newHV();
@@ -489,9 +513,9 @@ static int perl_instantiate(CONF_SECTION *conf, void **instance)
 	rad_request_proxy_reply_hv = newHV();
 
 	rad_reply_hv = get_hv("RAD_REPLY",1);
-        rad_check_hv = get_hv("RAD_CHECK",1);
+	rad_check_hv = get_hv("RAD_CHECK",1);
 	rad_config_hv = get_hv("RAD_CONFIG",1);
-        rad_request_hv = get_hv("RAD_REQUEST",1);
+	rad_request_hv = get_hv("RAD_REQUEST",1);
 	rad_request_proxy_hv = get_hv("RAD_REQUEST_PROXY",1);
 	rad_request_proxy_reply_hv = get_hv("RAD_REQUEST_PROXY_REPLY",1);
 
@@ -516,17 +540,20 @@ static int perl_instantiate(CONF_SECTION *conf, void **instance)
  */
 static void perl_store_vps(VALUE_PAIR *vp, HV *rad_hv)
 {
-        VALUE_PAIR	*nvp, *vpa, *vpn;
+	VALUE_PAIR	*nvp, *vpa, *vpn;
 	AV		*av;
-	char            buffer[1024];
+	char		namebuf[256], *name;
+	char	    buffer[1024];
 	int		attr, len;
 
 	hv_undef(rad_hv);
 	nvp = paircopy(vp);
 
 	while (nvp != NULL) {
+		name = nvp->name;
 		attr = nvp->attribute;
 		vpa = paircopy2(nvp,attr);
+
 		if (vpa->next) {
 			av = newAV();
 			vpn = vpa;
@@ -539,10 +566,17 @@ static void perl_store_vps(VALUE_PAIR *vp, HV *rad_hv)
 			hv_store(rad_hv, nvp->name, strlen(nvp->name),
 					newRV_noinc((SV *) av), 0);
 		} else {
+			if ((vpa->flags.has_tag) &&
+			    (vpa->flags.tag != 0)) {
+				snprintf(namebuf, sizeof(namebuf), "%s:%d",
+					 nvp->name, nvp->flags.tag);
+				name = namebuf;
+			}
+
 			len = vp_prints_value(buffer, sizeof(buffer),
-					vpa, FALSE);
-			hv_store(rad_hv, vpa->name, strlen(vpa->name),
-					newSVpv(buffer, len), 0);
+					      vpa, FALSE);
+			hv_store(rad_hv, name, strlen(name),
+				 newSVpv(buffer, len), 0);
 		}
 
 		pairfree(&vpa);
@@ -560,23 +594,23 @@ static void perl_store_vps(VALUE_PAIR *vp, HV *rad_hv)
  *
  */
 static int pairadd_sv(VALUE_PAIR **vp, char *key, SV *sv, int operator) {
-       char            *val;
+       char	    *val;
        VALUE_PAIR      *vpp;
 
        if (SvOK(sv)) {
-               val = SvPV_nolen(sv);
-               vpp = pairmake(key, val, operator);
-               if (vpp != NULL) {
-                       pairadd(vp, vpp);
-                       radlog(L_DBG,
-                         "rlm_perl: Added pair %s = %s", key, val);
+	       val = SvPV_nolen(sv);
+	       vpp = pairmake(key, val, operator);
+	       if (vpp != NULL) {
+		       pairadd(vp, vpp);
+		       radlog(L_DBG,
+			 "rlm_perl: Added pair %s = %s", key, val);
 		       return 1;
-               } else {
-                       radlog(L_DBG,
-                         "rlm_perl: ERROR: Failed to create pair %s = %s",
-                         key, val);
-               }
-        }
+	       } else {
+		       radlog(L_DBG,
+			 "rlm_perl: ERROR: Failed to create pair %s = %s",
+			 key, val);
+	       }
+	}
        return 0;
 }
 
@@ -594,18 +628,18 @@ static int get_hv_content(HV *my_hv, VALUE_PAIR **vp)
 
        *vp = NULL;
        for (i = hv_iterinit(my_hv); i > 0; i--) {
-               res_sv = hv_iternextsv(my_hv,&key,&key_len);
-               if (SvROK(res_sv) && (SvTYPE(SvRV(res_sv)) == SVt_PVAV)) {
-                       av = (AV*)SvRV(res_sv);
-                       len = av_len(av);
-                       for (j = 0; j <= len; j++) {
-                               av_sv = av_fetch(av, j, 0);
-                               ret = pairadd_sv(vp, key, *av_sv, T_OP_ADD) + ret;
-                       }
-               } else ret = pairadd_sv(vp, key, res_sv, T_OP_EQ) + ret;
-        }
+	       res_sv = hv_iternextsv(my_hv,&key,&key_len);
+	       if (SvROK(res_sv) && (SvTYPE(SvRV(res_sv)) == SVt_PVAV)) {
+		       av = (AV*)SvRV(res_sv);
+		       len = av_len(av);
+		       for (j = 0; j <= len; j++) {
+			       av_sv = av_fetch(av, j, 0);
+			       ret = pairadd_sv(vp, key, *av_sv, T_OP_ADD) + ret;
+		       }
+	       } else ret = pairadd_sv(vp, key, res_sv, T_OP_EQ) + ret;
+	}
 
-        return ret;
+	return ret;
 }
 
 /*
@@ -629,16 +663,21 @@ static int rlmperl_call(void *instance, REQUEST *request, char *function_name)
 	HV		*rad_request_proxy_reply_hv;
 
 #ifdef USE_ITHREADS
+	pthread_mutex_lock(&inst->clone_mutex);
+
 	PerlInterpreter *interp;
 
-	interp = rlm_perl_clone(inst->perl);
+	interp = rlm_perl_clone(inst->perl,inst->thread_key);
 	{
 	  dTHXa(interp);
 	  PERL_SET_CONTEXT(interp);
 	}
+
+	pthread_mutex_unlock(&inst->clone_mutex);
 #else
 	PERL_SET_CONTEXT(inst->perl);
 #endif
+
 	{
 	dSP;
 
@@ -797,10 +836,10 @@ static int perl_accounting(void *instance, REQUEST *request)
 
 	if ((pair = pairfind(request->packet->vps, PW_ACCT_STATUS_TYPE)) != NULL) {
 		acctstatustype = pair->vp_integer;
-        } else {
-                radlog(L_ERR, "Invalid Accounting Packet");
-                return RLM_MODULE_INVALID;
-        }
+	} else {
+		radlog(L_ERR, "Invalid Accounting Packet");
+		return RLM_MODULE_INVALID;
+	}
 
 	switch (acctstatustype) {
 
@@ -863,6 +902,24 @@ static int perl_post_auth(void *instance, REQUEST *request)
 	return rlmperl_call(instance, request,
 			((PERL_INST *)instance)->func_post_auth);
 }
+#ifdef WITH_COA
+/*
+ *	Recv CoA request
+ */
+static int perl_recv_coa(void *instance, REQUEST *request)
+{
+	return rlmperl_call(instance, request,
+			((PERL_INST *)instance)->func_recv_coa);
+}
+/*
+ *	Send CoA request
+ */
+static int perl_send_coa(void *instance, REQUEST *request)
+{
+	return rlmperl_call(instance, request,
+			((PERL_INST *)instance)->func_send_coa);
+}
+#endif
 /*
  * Detach a instance give a chance to a module to make some internal setup ...
  */
@@ -875,63 +932,65 @@ static int perl_detach(void *instance)
 	/*
 	 *	FIXME: Call this in the destruct function?
 	 */
-		{
+	{
 		dTHXa(handle->clone);
 		PERL_SET_CONTEXT(handle->clone);
 		{
-		dSP; ENTER; SAVETMPS; PUSHMARK(SP);
-		count = call_pv(inst->func_detach, G_SCALAR | G_EVAL );
-		SPAGAIN;
+			dSP; ENTER; SAVETMPS; PUSHMARK(SP);
+			count = call_pv(inst->func_detach, G_SCALAR | G_EVAL );
+			SPAGAIN;
 
-		if (count == 1) {
-			exitstatus = POPi;
-			/*
-			 * FIXME: bug in perl
-			 *
-			 */
-			if (exitstatus >= 100 || exitstatus < 0) {
-				exitstatus = RLM_MODULE_FAIL;
+			if (count == 1) {
+				exitstatus = POPi;
+				/*
+				 * FIXME: bug in perl
+				 *
+				 */
+				if (exitstatus >= 100 || exitstatus < 0) {
+					exitstatus = RLM_MODULE_FAIL;
+				}
 			}
+			PUTBACK;
+			FREETMPS;
+			LEAVE;
 		}
-		PUTBACK;
-		FREETMPS;
-		LEAVE;
-		}
-		}
+	}
 #endif
 
-	{
-	dTHXa(inst->perl);
-	PERL_SET_CONTEXT(inst->perl);
-	{
-	dSP; ENTER; SAVETMPS;
-	PUSHMARK(SP);
+	if (inst->func_detach) {
+		dTHXa(inst->perl);
+		PERL_SET_CONTEXT(inst->perl);
+		{
+			dSP; ENTER; SAVETMPS;
+			PUSHMARK(SP);
 
-	count = call_pv(inst->func_detach, G_SCALAR | G_EVAL );
-	SPAGAIN;
+			count = call_pv(inst->func_detach, G_SCALAR | G_EVAL );
+			SPAGAIN;
 
-	if (count == 1) {
-		exitstatus = POPi;
-		if (exitstatus >= 100 || exitstatus < 0) {
-			exitstatus = RLM_MODULE_FAIL;
+			if (count == 1) {
+				exitstatus = POPi;
+				if (exitstatus >= 100 || exitstatus < 0) {
+					exitstatus = RLM_MODULE_FAIL;
+				}
+			}
+			PUTBACK;
+			FREETMPS;
+			LEAVE;
 		}
 	}
-	PUTBACK;
-	FREETMPS;
-	LEAVE;
-	}
-	}
 
-	xlat_unregister(inst->xlat_name, perl_xlat);
+	xlat_unregister(inst->xlat_name, perl_xlat, instance);
 	free(inst->xlat_name);
 
 #ifdef USE_ITHREADS
 	rlm_perl_destruct(inst->perl);
+	pthread_mutex_destroy(&inst->clone_mutex);
 #else
 	perl_destruct(inst->perl);
 	perl_free(inst->perl);
 #endif
 
+	PERL_SYS_TERM();
 	free(inst);
 	return exitstatus;
 }
@@ -965,5 +1024,9 @@ module_t rlm_perl = {
 		perl_pre_proxy,		/* pre-proxy */
 		perl_post_proxy,	/* post-proxy */
 		perl_post_auth		/* post-auth */
+#ifdef WITH_COA
+		, perl_recv_coa,
+		perl_send_coa
+#endif
 	},
 };
