@@ -2,7 +2,7 @@
  * rlm_sql.c		SQL Module
  * 		Main SQL module file. Most ICRADIUS code is located in sql.c
  *
- * Version:	$Id$
+ * Version:	$Id: 056ca05b067d895c6b1288401ff0d7a4d6d57499 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -24,12 +24,11 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: 056ca05b067d895c6b1288401ff0d7a4d6d57499 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
 #include <freeradius-devel/rad_assert.h>
-#include <ltdl.h>
 
 #include <sys/stat.h>
 
@@ -50,6 +49,8 @@ static const CONF_PARSER module_config[] = {
 	 offsetof(SQL_CONFIG,sql_password), NULL, ""},
 	{"radius_db", PW_TYPE_STRING_PTR,
 	 offsetof(SQL_CONFIG,sql_db), NULL, "radius"},
+	{"filename", PW_TYPE_FILENAME, /* for sqlite */
+	 offsetof(SQL_CONFIG,sql_file), NULL, NULL},
 	{"read_groups", PW_TYPE_BOOLEAN,
 	 offsetof(SQL_CONFIG,read_groups), NULL, "yes"},
 	{"sqltrace", PW_TYPE_BOOLEAN,
@@ -62,6 +63,10 @@ static const CONF_PARSER module_config[] = {
 	 offsetof(SQL_CONFIG,deletestalesessions), NULL, "yes"},
 	{"num_sql_socks", PW_TYPE_INTEGER,
 	 offsetof(SQL_CONFIG,num_sql_socks), NULL, "5"},
+	{"lifetime", PW_TYPE_INTEGER,
+	 offsetof(SQL_CONFIG,lifetime), NULL, "0"},
+	{"max_queries", PW_TYPE_INTEGER,
+	 offsetof(SQL_CONFIG,max_queries), NULL, "0"},
 	{"sql_user_name", PW_TYPE_STRING_PTR,
 	 offsetof(SQL_CONFIG,query_user), NULL, ""},
 	{"default_user_profile", PW_TYPE_STRING_PTR,
@@ -104,6 +109,12 @@ static const CONF_PARSER module_config[] = {
 	 offsetof(SQL_CONFIG,allowed_chars), NULL,
 	"@abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-_: /"},
 
+	/*
+	 *	This only works for a few drivers.
+	 */
+	{"query_timeout", PW_TYPE_INTEGER,
+	 offsetof(SQL_CONFIG,query_timeout), NULL, NULL},
+	 
 	{NULL, -1, 0, NULL, NULL}
 };
 
@@ -129,6 +140,9 @@ static size_t sql_escape_func(char *out, size_t outlen, const char *in);
 /*
  *	sql xlat function. Right now only SELECTs are supported. Only
  *	the first element of the SELECT result will be used.
+ *
+ *	For other statements (insert, update, delete, etc.), the
+ *	number of affected rows will be returned.
  */
 static int sql_xlat(void *instance, REQUEST *request,
 		    char *fmt, char *out, size_t freespace,
@@ -162,6 +176,59 @@ static int sql_xlat(void *instance, REQUEST *request,
 	sqlsocket = sql_get_socket(inst);
 	if (sqlsocket == NULL)
 		return 0;
+
+	/*
+	 *	If the query starts with any of the following prefixes,
+	 *	then return the number of rows affected
+	 */
+	if ((strncasecmp(querystr, "insert", 6) == 0) ||
+	    (strncasecmp(querystr, "update", 6) == 0) ||
+	    (strncasecmp(querystr, "delete", 6) == 0)) {
+		int numaffected;
+		char buffer[21]; /* 64bit max is 20 decimal chars + null byte */
+
+		if (rlm_sql_query(sqlsocket,inst,querystr)) {
+			radlog(L_ERR, "rlm_sql (%s): database query error, %s: %s",
+				inst->config->xlat_name, querystr,
+				(inst->module->sql_error)(sqlsocket,
+							  inst->config));
+			sql_release_socket(inst,sqlsocket);
+			return 0;
+		}
+	       
+		numaffected = (inst->module->sql_affected_rows)(sqlsocket,
+								inst->config);
+		if (numaffected < 1) {
+			RDEBUG("rlm_sql (%s): SQL query affected no rows",
+				inst->config->xlat_name);
+		}
+
+		/*
+		 *	Don't chop the returned number if freespace is
+		 *	too small.  This hack is necessary because
+		 *	some implementations of snprintf return the
+		 *	size of the written data, and others return
+		 *	the size of the data they *would* have written
+		 *	if the output buffer was large enough.
+		 */
+		snprintf(buffer, sizeof(buffer), "%d", numaffected);
+		ret = strlen(buffer);
+		if (ret >= freespace){
+			RDEBUG("rlm_sql (%s): Can't write result, insufficient string space",
+			       inst->config->xlat_name);
+			(inst->module->sql_finish_query)(sqlsocket,
+							 inst->config);
+			sql_release_socket(inst,sqlsocket);
+			return 0;
+		}
+		
+		memcpy(out, buffer, ret + 1); /* we did bounds checking above */
+
+		(inst->module->sql_finish_query)(sqlsocket, inst->config);
+		sql_release_socket(inst,sqlsocket);
+		return ret;
+	} /* else it's a SELECT statement */
+
 	if (rlm_sql_select_query(sqlsocket,inst,querystr)){
 		radlog(L_ERR, "rlm_sql (%s): database query error, %s: %s",
 		       inst->config->xlat_name,querystr,
@@ -333,12 +400,13 @@ static int generate_sql_clients(SQL_INST *inst)
 			c->nastype = strdup(row[3]);
 
 		numf = (inst->module->sql_num_fields)(sqlsocket, inst->config);
-		if ((numf > 5) && (row[5] != NULL)) c->server = strdup(row[5]);
+		if ((numf > 5) && (row[5] != NULL) && *row[5]) c->server = strdup(row[5]);
 
 		DEBUG("rlm_sql (%s): Adding client %s (%s, server=%s) to clients list",
 		      inst->config->xlat_name,
 		      c->longname,c->shortname, c->server ? c->server : "<none>");
 		if (!client_add(NULL, c)) {
+			sql_release_socket(inst, sqlsocket);
 			DEBUG("rlm_sql (%s): Failed to add client %s (%s) to clients list.  Maybe there's a duplicate?",
 			      inst->config->xlat_name,
 			      c->longname,c->shortname);
@@ -361,6 +429,24 @@ static size_t sql_escape_func(char *out, size_t outlen, const char *in)
 	size_t len = 0;
 
 	while (in[0]) {
+		size_t utf8_len;
+
+		/*
+		 *	Allow all multi-byte UTF8 characters.
+		 */
+		utf8_len = fr_utf8_char((uint8_t const *) in);
+		if (utf8_len > 1) {
+			if (outlen <= utf8_len) break;
+
+			memcpy(out, in, utf8_len);
+			in += utf8_len;
+			out += utf8_len;
+
+			outlen -= utf8_len;
+			len += utf8_len;
+			continue;
+		}
+
 		/*
 		 *	Non-printable characters get replaced with their
 		 *	mime-encoded equivalents.
@@ -500,6 +586,7 @@ static int sql_get_grouplist (SQL_INST *inst, SQLSOCK *sqlsocket, REQUEST *reque
 			*group_list = rad_malloc(sizeof(SQL_GROUPLIST));
 			group_list_tmp = *group_list;
 		} else {
+			rad_assert(group_list_tmp != NULL);
 			group_list_tmp->next = rad_malloc(sizeof(SQL_GROUPLIST));
 			group_list_tmp = group_list_tmp->next;
 		}
@@ -623,6 +710,7 @@ static int rlm_sql_process_groups(SQL_INST *inst, REQUEST *request, SQLSOCK *sql
 		if (!sql_group) {
 			radlog_request(L_ERR, 0, request,
 				       "Error creating Sql-Group attribute");
+			sql_grouplist_free(&group_list);
 			return -1;
 		}
 		pairadd(&request->packet->vps, sql_group);
@@ -631,6 +719,7 @@ static int rlm_sql_process_groups(SQL_INST *inst, REQUEST *request, SQLSOCK *sql
 				       "Error generating query; rejecting user");
 			/* Remove the grouup we added above */
 			pairdelete(&request->packet->vps, PW_SQL_GROUP);
+			sql_grouplist_free(&group_list);
 			return -1;
 		}
 		rows = sql_getvpdata(inst, sqlsocket, &check_tmp, querystr);
@@ -640,6 +729,7 @@ static int rlm_sql_process_groups(SQL_INST *inst, REQUEST *request, SQLSOCK *sql
 			/* Remove the grouup we added above */
 			pairdelete(&request->packet->vps, PW_SQL_GROUP);
 			pairfree(&check_tmp);
+			sql_grouplist_free(&group_list);
 			return -1;
 		} else if (rows > 0) {
 			/*
@@ -657,6 +747,7 @@ static int rlm_sql_process_groups(SQL_INST *inst, REQUEST *request, SQLSOCK *sql
 					/* Remove the grouup we added above */
 					pairdelete(&request->packet->vps, PW_SQL_GROUP);
 					pairfree(&check_tmp);
+					sql_grouplist_free(&group_list);
 					return -1;
 				}
 				if (sql_getvpdata(inst, sqlsocket, &reply_tmp, querystr) < 0) {
@@ -666,6 +757,7 @@ static int rlm_sql_process_groups(SQL_INST *inst, REQUEST *request, SQLSOCK *sql
 					pairdelete(&request->packet->vps, PW_SQL_GROUP);
 					pairfree(&check_tmp);
 					pairfree(&reply_tmp);
+					sql_grouplist_free(&group_list);
 					return -1;
 				}
 				*dofallthrough = fallthrough(reply_tmp);
@@ -690,6 +782,7 @@ static int rlm_sql_process_groups(SQL_INST *inst, REQUEST *request, SQLSOCK *sql
 				/* Remove the grouup we added above */
 				pairdelete(&request->packet->vps, PW_SQL_GROUP);
 				pairfree(&check_tmp);
+				sql_grouplist_free(&group_list);
 				return -1;
 			}
 			if (sql_getvpdata(inst, sqlsocket, &reply_tmp, querystr) < 0) {
@@ -699,6 +792,7 @@ static int rlm_sql_process_groups(SQL_INST *inst, REQUEST *request, SQLSOCK *sql
 				pairdelete(&request->packet->vps, PW_SQL_GROUP);
 				pairfree(&check_tmp);
 				pairfree(&reply_tmp);
+				sql_grouplist_free(&group_list);
 				return -1;
 			}
 			*dofallthrough = fallthrough(reply_tmp);
@@ -734,35 +828,16 @@ static int rlm_sql_detach(void *instance)
 		}
 
 		if (inst->config->xlat_name) {
-			xlat_unregister(inst->config->xlat_name,(RAD_XLAT_FUNC)sql_xlat);
+			xlat_unregister(inst->config->xlat_name,(RAD_XLAT_FUNC)sql_xlat, instance);
 			free(inst->config->xlat_name);
 		}
 
 		/*
-		 *	Free up dynamically allocated string pointers.
+		 *	Catch multiple instances of the module.
 		 */
-		for (i = 0; module_config[i].name != NULL; i++) {
-			char **p;
-			if (module_config[i].type != PW_TYPE_STRING_PTR) {
-				continue;
-			}
-
-			/*
-			 *	Treat 'config' as an opaque array of bytes,
-			 *	and take the offset into it.  There's a
-			 *      (char*) pointer at that offset, and we want
-			 *	to point to it.
-			 */
-			p = (char **) (((char *)inst->config) + module_config[i].offset);
-			if (!*p) { /* nothing allocated */
-				continue;
-			}
-			free(*p);
-			*p = NULL;
+		if (allowed_chars == inst->config->allowed_chars) {
+			allowed_chars = NULL;
 		}
-		allowed_chars = NULL;
-		free(inst->config);
-		inst->config = NULL;
 	}
 
 	if (inst->handle) {
@@ -779,15 +854,20 @@ static int rlm_sql_detach(void *instance)
 }
 static int rlm_sql_instantiate(CONF_SECTION * conf, void **instance)
 {
+	int i;
 	SQL_INST *inst;
 	const char *xlat_name;
 
 	inst = rad_malloc(sizeof(SQL_INST));
 	memset(inst, 0, sizeof(SQL_INST));
 
-	inst->config = rad_malloc(sizeof(SQL_CONFIG));
-	memset(inst->config, 0, sizeof(SQL_CONFIG));
-
+	/*
+	 *	The server core expects to be able to do
+	 *	cf_section_parse_free(inst).  So this hack is
+	 *	necessary.
+	 */
+	inst->config = &inst->myconfig;
+	
 	/*
 	 *	Export these methods, too.  This avoids RTDL_GLOBAL.
 	 */
@@ -795,6 +875,9 @@ static int rlm_sql_instantiate(CONF_SECTION * conf, void **instance)
 	inst->sql_get_socket = sql_get_socket;
 	inst->sql_release_socket = sql_release_socket;
 	inst->sql_escape_func = sql_escape_func;
+	inst->sql_query = rlm_sql_query;
+	inst->sql_select_query = rlm_sql_select_query;
+	inst->sql_fetch_row = rlm_sql_fetch_row;
 
 	/*
 	 * If the configuration parameters can't be parsed, then
@@ -806,8 +889,38 @@ static int rlm_sql_instantiate(CONF_SECTION * conf, void **instance)
 	}
 
 	xlat_name = cf_section_name2(conf);
-	if (xlat_name == NULL)
+	if (xlat_name == NULL) {
 		xlat_name = cf_section_name1(conf);
+	} else {
+		char *group_name;
+		DICT_ATTR *dattr;
+		ATTR_FLAGS flags;
+
+		/*
+		 * Allocate room for <instance>-SQL-Group
+		 */
+		group_name = rad_malloc((strlen(xlat_name) + 1 + 11) * sizeof(char));
+		sprintf(group_name,"%s-SQL-Group",xlat_name);
+		DEBUG("rlm_sql Creating new attribute %s",group_name);
+
+		memset(&flags, 0, sizeof(flags));
+		dict_addattr(group_name, 0, PW_TYPE_STRING, -1, flags);
+		dattr = dict_attrbyname(group_name);
+		if (dattr == NULL){
+			radlog(L_ERR, "rlm_sql: Failed to create attribute %s",group_name);
+			free(group_name);
+			free(inst);	/* FIXME: detach */
+			return -1;
+		}
+
+		if (inst->config->groupmemb_query && 
+		    inst->config->groupmemb_query[0]) {
+			DEBUG("rlm_sql: Registering sql_groupcmp for %s",group_name);
+			paircompare_register(dattr->attr, PW_USER_NAME, sql_groupcmp, inst);
+		}
+
+		free(group_name);
+	}
 	if (xlat_name){
 		inst->config->xlat_name = strdup(xlat_name);
 		xlat_register(xlat_name, (RAD_XLAT_FUNC)sql_xlat, inst);
@@ -862,7 +975,10 @@ static int rlm_sql_instantiate(CONF_SECTION * conf, void **instance)
 		return -1;
 	}
 
-	paircompare_register(PW_SQL_GROUP, PW_USER_NAME, sql_groupcmp, inst);
+	if (inst->config->groupmemb_query && 
+	    inst->config->groupmemb_query[0]) {
+		paircompare_register(PW_SQL_GROUP, PW_USER_NAME, sql_groupcmp, inst);
+	}
 
 	if (inst->config->do_clients){
 		if (generate_sql_clients(inst) == -1){
@@ -873,6 +989,23 @@ static int rlm_sql_instantiate(CONF_SECTION * conf, void **instance)
 	}
 	allowed_chars = inst->config->allowed_chars;
 
+	for (i = 0; module_config[i].name != NULL; i++) {
+		char **p;
+
+		if (module_config[i].type != PW_TYPE_STRING_PTR) continue;
+		if (strstr(module_config[i].name, "_query") != NULL) continue;
+		
+		p = (char **) (((char *)inst->config) + module_config[i].offset);
+
+		if (!*p) continue;
+
+		if (strlen(*p) > ((2 * MAX_QUERY_LEN) / 3)) {
+			DEBUG("%s: WARNING Query '%s' is probably too long!",
+			      inst->config->xlat_name, module_config[i].name);
+		}
+	}
+
+
 	*instance = inst;
 
 	return RLM_MODULE_OK;
@@ -1109,6 +1242,10 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 					       (inst->module->sql_error)(sqlsocket, inst->config));
 					ret = RLM_MODULE_FAIL;
 				}
+				/*
+				 *	If no one is online, num_affected_rows
+				 *	will be zero, which is OK.
+				 */
 				(inst->module->sql_finish_query)(sqlsocket, inst->config);
 			}
 
@@ -1153,8 +1290,12 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 								radlog_request(L_ERR, 0, request, "Couldn't insert SQL accounting ALIVE record - %s",
 								       (inst->module->sql_error)(sqlsocket, inst->config));
 								ret = RLM_MODULE_FAIL;
+							} else {
+								numaffected = (inst->module->sql_affected_rows)(sqlsocket, inst->config);
+								if (numaffected < 1) {
+									ret = RLM_MODULE_NOOP;
+								}
 							}
-							(inst->module->sql_finish_query)(sqlsocket, inst->config);
 						}
 					}
 				}
@@ -1197,7 +1338,11 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 							       (inst->module->sql_error)(sqlsocket, inst->config));
 							ret = RLM_MODULE_FAIL;
 						}
-						(inst->module->sql_finish_query)(sqlsocket, inst->config);
+					}
+				} else {
+					numaffected = (inst->module->sql_affected_rows)(sqlsocket, inst->config);
+					if (numaffected < 1) {
+						ret = RLM_MODULE_NOOP;
 					}
 				}
 				(inst->module->sql_finish_query)(sqlsocket, inst->config);
@@ -1247,9 +1392,9 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 
 						if (acctsessiontime <= 0) {
 							radius_xlat(logstr, sizeof(logstr), "stop packet with zero session length. [user '%{User-Name}', nas '%{NAS-IP-Address}']", request, NULL);
-							radlog_request(L_ERR, 0, request, "%s", logstr);
+							radlog_request(L_DBG, 0, request, "%s", logstr);
 							sql_release_socket(inst, sqlsocket);
-							ret = RLM_MODULE_NOOP;
+							return RLM_MODULE_NOOP;
 						}
 #endif
 
@@ -1262,8 +1407,12 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 
 								       (inst->module->sql_error)(sqlsocket, inst->config));
 								ret = RLM_MODULE_FAIL;
+							} else {
+								numaffected = (inst->module->sql_affected_rows)(sqlsocket, inst->config);
+								if (numaffected < 1) {
+									ret = RLM_MODULE_NOOP;
+								}
 							}
-							(inst->module->sql_finish_query)(sqlsocket, inst->config);
 						}
 					}
 				}
